from __future__ import with_statement
import re
from .diagnose_error import get_coq_output
from .coq_version import get_coq_native_compiler_ondemand_fragment, get_boot_noinputstate_args, coqlib_args_of_coq_args
from .coq_full_grammar import COQ_GRAMMAR_TOKENS

# Like coq_version.py, except for things that use get_coq_output (or
# perhaps a bit more generally for things that need to run coqc on a
# file)

__all__ = ["get_proof_term_works_with_time", "get_ltac_support_snippet", "get_reserved_modnames"]


def get_proof_term_works_with_time(coqc_prog, **kwargs):
    contents = r"""Lemma foo : forall _ : Type, Type.
Proof (fun x => x)."""
    output, cmds, retcode, runtime = get_coq_output(
        coqc_prog,
        ("-time", "-q", *get_boot_noinputstate_args(coqc_prog, **kwargs)),
        contents,
        timeout_val=1,
        verbose_base=3,
        **kwargs,
    )
    return "Error: Attempt to save an incomplete proof" not in output


LTAC_SUPPORT_SNIPPET = {}


def get_ltac_support_snippet(coqc, coqc_args=(), **kwargs):
    if coqc in LTAC_SUPPORT_SNIPPET.keys():
        return LTAC_SUPPORT_SNIPPET[coqc]
    test = r"""Inductive False : Prop := .
Axiom proof_admitted : False.
Tactic Notation "admit" := abstract case proof_admitted.
Goal False. admit. Qed."""
    errinfo = {}
    native_ondemand_args = list(get_coq_native_compiler_ondemand_fragment(coqc, coq_args=coqc_args, **kwargs))
    for before, after in (
        ("Require Coq.Init.Ltac.\n", "Import Coq.Init.Ltac.\n"),
        ("Require Coq.Init.Notations.\n", "Import Coq.Init.Notations.\n"),
        ("Require Stdlib.Init.Ltac.\n", "Import Stdlib.Init.Ltac.\n"),
        ("Require Stdlib.Init.Notations.\n", "Import Stdlib.Init.Notations.\n"),
        ('Declare ML Module "coq-core.plugins.ltac".\n', ""),
        ('Declare ML Module "coq-core.plugins.ltac".\n', 'Global Set Default Proof Mode "Classic".\n'),
        ('Declare ML Module "ltac_plugin".\n', ""),
        ('Declare ML Module "ltac_plugin".\n', 'Global Set Default Proof Mode "Classic".\n'),
    ):
        contents = "%s\n%s\n%s" % (before, after, test)
        output, cmds, retcode, runtime = get_coq_output(
            coqc,
            tuple(["-q", "-nois", *coqlib_args_of_coq_args(coqc, coqc_args, **kwargs)] + native_ondemand_args),
            contents,
            timeout_val=None,
            verbose_base=3,
            is_coqtop=kwargs["coqc_is_coqtop"],
            **kwargs,
        )
        if retcode == 0:
            LTAC_SUPPORT_SNIPPET[coqc] = (before, after)
            return (before, after)
        else:
            errinfo[contents] = {"output": output, "cmds": cmds, "retcode": retcode, "runtime": runtime}
    raise Exception("No valid ltac support snipped found.  Debugging info: %s" % repr(errinfo))


def get_is_modname_valid(coqc_prog, modname, **kwargs):
    contents = "Module %s. End %s." % (modname, modname)
    if " " in modname:
        return False
    output, cmds, retcode, runtime = get_coq_output(
        coqc_prog,
        ("-q", *get_boot_noinputstate_args(coqc_prog, **kwargs)),
        contents,
        timeout_val=1,
        verbose_base=3,
        **kwargs,
    )
    return "Syntax error:" not in output


def get_reserved_modnames(coqtop_prog, **kwargs):
    grammars_contents = "Print Grammar tactic. Print Grammar constr. Print Grammar vernac."
    grammars_output, cmds, retcode, runtime = get_coq_output(
        coqtop_prog,
        ("-q", *get_boot_noinputstate_args(coqtop_prog, **kwargs)),
        grammars_contents,
        is_coqtop=True,
        pass_on_stdin=True,
        timeout_val=1,
        verbose_base=3,
        **kwargs,
    )
    tokens = sorted(set([i.strip('"') for i in re.findall(r'"[a-zA-Z_][^"]*"', grammars_output)] + COQ_GRAMMAR_TOKENS))
    contents = "\n".join("Module %s. End %s." % (modname, modname) for modname in tokens)
    output, cmds, retcode, runtime = get_coq_output(
        coqtop_prog,
        ("-q", *get_boot_noinputstate_args(coqtop_prog, **kwargs)),
        contents,
        is_coqtop=True,
        pass_on_stdin=True,
        timeout_val=10,
        verbose_base=3,
        **kwargs,
    )
    success = re.findall(r"Module ([^ ]+) is defined", output)
    return tuple(modname for modname in tokens if modname not in success)



import os, time
from .memoize import memoize
from . import util

__all__ = [
    "clean_v_file",
    "clean_extra_coq_files",
    "write_bytes_to_file",
    "read_bytes_from_file",
    "write_to_file",
    "read_from_file",
    "restore_file",
]


def clean_extra_coq_files(v_file_name, extra_exts=tuple()):
    for pre in ("", "."):
        for ext in tuple(list(extra_exts) + [".glob", ".vo", ".d", ".v.d", ".aux", ".vos", ".vok"]):
            name = "".join((os.path.dirname(v_file_name[:-2]), os.sep, pre, os.path.basename(v_file_name[:-2]), ext))
            if os.path.exists(name):
                os.remove(name)


def clean_v_file(file_name):
    clean_extra_coq_files(file_name, extra_exts=(".v",))


def backup(file_name, ext=".bak"):
    if not ext:
        raise ValueError
    if os.path.exists(file_name):
        backup(file_name + ext, ext=ext)
        os.rename(file_name, file_name + ext)


@memoize
def memobackup(file_name, ext=".bak"):
    return backup(file_name, ext)


FILE_CACHE = {}


def write_bytes_to_file(file_name, contents, do_backup=False, backup_ext=".bak", memoize=False):
    assert contents is bytes(contents)
    backed_up = False
    while not backed_up:
        try:
            if do_backup:
                if memoize:
                    memobackup(file_name, ext=backup_ext)
                else:
                    backup(file_name, ext=backup_ext)
            backed_up = True
        except IOError as e:
            print("Warning: f.write(%s) failed with %s\nTrying again in 10s" % (file_name, repr(e)))
            time.sleep(10)
    written = False
    while not written:
        try:
            with open(file_name, "wb") as f:
                f.write(contents)
            written = True
            FILE_CACHE[file_name] = (os.stat(file_name).st_mtime, contents)
        except IOError as e:
            print("Warning: f.write(%s) failed with %s\nTrying again in 10s" % (file_name, repr(e)))
            time.sleep(10)


def write_to_file(file_name, contents, *args, **kwargs):
    return write_bytes_to_file(file_name, contents.replace("\n", os.linesep).encode("utf-8"), *args, **kwargs)


def restore_file(file_name, backup_ext=".bak", backup_backup_ext=".unbak"):
    if not os.path.exists(file_name + backup_ext):
        raise IOError
    if os.path.exists(file_name):
        if backup_backup_ext:
            backup(file_name, backup_backup_ext)
        else:
            os.remove(file_name)
    os.rename(file_name + backup_ext, file_name)


def read_bytes_from_file(file_name):
    if file_name in FILE_CACHE.keys() and os.stat(file_name).st_mtime == FILE_CACHE[file_name][0]:
        return FILE_CACHE[file_name][1]
    with open(file_name, "rb") as f:
        FILE_CACHE[file_name] = (os.stat(file_name).st_mtime, f.read())
        return FILE_CACHE[file_name][1]


def read_from_file(file_name):
    return util.normalize_newlines(read_bytes_from_file(file_name).decode("utf-8"))
